"use strict";

var webpack = require('webpack');
var fs = require('fs');
var lodash = require('lodash');
var tmp = require('tmp');
var nodeExternals = require('webpack-node-externals');
var jsonfile = require('jsonfile');
var path = require('path');
var cronparser = require('cron-parser');
var debug = require('debug')('compile');
var cwd = process.cwd();
var whitelist = {
  '@grpc/proto-loader': 1,
  'ajv': 1,
  'async': 1,
  '@aws-sdk/client-s3': 1,
  'axios': 1,
  'busboy': 1,
  'crc-32': 1,
  'cron-parser': 1,
  'debug': 1,
  'fast-sort': 1,
  'form-data': 1,
  'google-auth-library': 1,
  'googleapis': 1,
  'grpc': 1,
  'jsonwebtoken': 1,
  'lodash': 1,
  'mongodb': 1,
  'mongodb-2': 1,
  'mongodb-3': 1,
  'mongodb-4': 1,
  'mongodb-5': 1,
  'nanoid': 1,
  'node-cron': 1,
  'node-fetch': 1,
  'node-geocoder': 1,
  'openid-client': 1,
  'passport-jwt': 1,
  'pdfkit': 1,
  'query-to-mongo': 1,
  'remarkable': 1,
  'request': 1,
  'sift': 1,
  'stripe': 1,
  'xml2js': 1,
  'yup': 1
};
var custom_pkg = [];
var diff_pkg = [];
module.exports = function (compileDir) {
  var lang = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 'Javascript';
  var minify = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : true;
  var useWhitelist = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : true;
  return new Promise(function (resolve, reject) {
    var bIsTypescript = lang !== 'Javascript';
    var bMinimize = minify;
    var tmpobj = tmp.dirSync();
    var tmpbundle = tmpobj.name;
    //console.log("Compile", compileDir, lang, tmpbundle);
    webpack([{
      target: 'node',
      entry: "".concat(compileDir, "/index.").concat(bIsTypescript ? 'ts' : 'js'),
      externalsPresets: {
        node: true
      },
      externals: [function (_ref, callback) {
        var context = _ref.context,
          request = _ref.request;
        if (useWhitelist && whitelist[request]) {
          debug('exclude', request);
          // Externalize to a commonjs module using the request path
          return callback(null, 'commonjs ' + request);
        }
        debug('include', request, context);
        // Continue without externalizing the import
        callback();
      }],
      optimization: {
        minimize: bMinimize
      },
      mode: 'production',
      stats: 'minimal',
      output: {
        libraryTarget: 'var',
        library: 'EntryPoint',
        filename: 'bundle.js',
        path: tmpbundle
      },
      module: {
        rules: [{
          test: /\.txt$|\.html|\.htm|\.css|\.hbs/,
          type: 'asset/inline',
          generator: {
            dataUrl: function dataUrl(content) {
              content = content.toString();
              return content;
            }
          }
        }, {
          test: /\.tsx?$/,
          loader: require.resolve('ts-loader')
        }]
      },
      resolve: {
        extensions: ['.ts', '.js']
      }
    }], function (err, stats) {
      // Stats Object            
      if (err || stats.hasErrors()) {
        if (err) console.log(err);
        var output = stats.toJson();
        var errJson = [];
        if (output.errors || output.warnings) {
          lodash.each(output.errors, function (e) {
            if (e && e.replace) {
              e = e.split('\n').slice(0, 2).join('\n');
              errJson.push(e);
            } else {
              errJson.push('🤔 ' + e.message || 'Aiii, some unknown error occured');
            }
          });
        }
        reject(errJson.join('\n').replace('You may need an appropriate loader to handle this file type, currently no loaders are configured to process this file. See https://webpack.js.org/concepts#loaders', ''));
        tmp.setGracefulCleanup();
      } else {
        // ToDo: try to test if module runs locally first   
        //const rt = require(`${tmpbundle}/bundle.js`);
        //const hook = rt.code.default;             
        fs.readFile("".concat(tmpbundle, "/bundle.js"), "utf8", function (err, src) {
          // parse manifest?
          src += '\n// build: ' + new Date().toISOString();
          resolve(src);
          tmp.setGracefulCleanup();
        });
      }
    });
  });
};