"use strict";

function _typeof(o) { "@babel/helpers - typeof"; return _typeof = "function" == typeof Symbol && "symbol" == typeof Symbol.iterator ? function (o) { return typeof o; } : function (o) { return o && "function" == typeof Symbol && o.constructor === Symbol && o !== Symbol.prototype ? "symbol" : typeof o; }, _typeof(o); }
var emitter = require('events').EventEmitter;
var debug = require('debug')('mockdb');
var jsonfile = require('jsonfile');
var j = JSON.stringify;
var lodash = require('lodash');
var _require = require('nanoid'),
  customAlphabet = _require.customAlphabet;
var alphabet = '0123456789abcdefghijklmnopqrstuvwxyz';
var nanoid = customAlphabet(alphabet, 14);
var operators = require('../lib/operators');
var pickAndOmit = operators.pickAndOmit;
var sift = require('sift');
var _require2 = require('fast-sort'),
  sort = _require2.sort;
var util = require('util');
var async = require('async');
var collectionMap = new Map();
var dbPath = null;
// utils
function toHex(str) {
  return parseInt(str, 10).toString(16).padStart(6, "0");
}
function isObject(val) {
  if (val === null) {
    return false;
  }
  return typeof val === 'function' || _typeof(val) === 'object';
}
var getCollection = function getCollection(collname) {
  //debug(collectionMap)
  if (!collectionMap.has(collname)) {
    collectionMap.set(collname, {});
  }
  return collectionMap.get(collname);
};

// filter _ttl values
function filterTTL(obj, nowInMillis, collname, key) {
  if (obj && obj._archiveAt && obj._archiveAt < nowInMillis) {
    // send to delete queue
    delete getCollection(collname);
    return true;
  }
  return false;
}
var createIndexKey = function createIndexKey(range) {
  if (lodash.isNumber(range)) {
    //range = range.toFixed(5);
    if (range < 0) {
      range = lodash.padStart(Math.abs(range).toFixed(5), 32, '0');
    } else {
      range = '1' + lodash.padStart(range.toFixed(5), 31, '0');
    }
  }
  return range;
};
var _updateIndexes = function _updateIndexes(collection, idxFunc) {
  var metaColl = getCollection('_dbmetadata');
  lodash.each(metaColl['indexes'], function (acoll, collkey) {
    console.log("index coll", collkey);
    var idxarr = [];
    lodash.each(acoll, function (aindexfield, aindexkey) {
      console.log("index field", aindexkey, aindexfield);
      idxarr.push(aindexkey);
    });
    idxFunc({
      "collection": collkey,
      "index": idxarr
    });
  });
};
var updateIndexes = lodash.debounce(_updateIndexes, 1000, {
  'maxWait': 3000
});

/*
    Update document with special operator
*/
function applyUpdateOperator(document, vitem, cb) {
  var localCallback = function localCallback(e, s) {
    cb(e, s);
  };
  if (document['$inc']) {
    operators.applyIncOperator(document, vitem, localCallback);
  }
  if (document['$push']) {
    operators.applyPushOperator(document, vitem, localCallback);
  }
  if (document['$pull']) {
    operators.applyPullOperator(document, vitem, localCallback);
  }
  if (document['$pullAll']) {
    operators.applyPullAllOperator(document, vitem, localCallback);
  }
  if (document['$pop']) {
    operators.applyPopOperator(document, vitem, localCallback);
  }
  if (document['$set'] || vitem['$set']) {
    operators.applySetOperator(document, vitem, localCallback);
  }
  if (document['$addToSet']) {
    operators.applyAddToSetOperator(document, vitem, localCallback);
  }
  if (document['$unset']) {
    operators.applyUnSetOperator(document, vitem, localCallback);
  }
  if (document['$rename']) {
    operators.applyRenameOperator(document, vitem, localCallback);
  }
}

// load / sava json db
var loadedDB = false;
var loadDB = function loadDB() {
  if (loadedDB === true) return;
  try {
    var data = jsonfile.readFileSync(dbPath);
    collectionMap = new Map(data);
    debug("loadDB");
    loadedDB = true;
  } catch (ex) {
    debug(ex);
    _saveDB();
  }
};
var _saveDB = function _saveDB() {
  debug("saveDB", dbPath);
  var data = Array.from(collectionMap.entries());
  jsonfile.writeFile(dbPath, data, {
    spaces: 2,
    EOL: '\r\n'
  });
};
var saveDB = lodash.debounce(_saveDB, 1000, {
  'maxWait': 3000
});

// mockDB impl
var _DB = function _DB(dbpath, queueFunc) {
  dbPath = dbpath;
  loadDB();
  var queue = async.queue(function (task, callback) {
    debug("Incoming queue", task);
    //delete getCollection(`_queue_${task.topic}`)[task._id];
    queueFunc(task, function (err, result) {
      callback();
    });
  }, 1);
  var addToQueue = function addToQueue(topic, qitem) {
    //getCollection(`_queue_${topic}`)[qitem._id] = qitem;
    queue.push(qitem);
    return qitem;
  };
  function createQueuePayload(key, topic, data, jobId, hints) {
    var _id = key;
    var payload = pickAndOmit(data, hints);
    var qitem = {
      _id: _id,
      topic: topic,
      payload: payload,
      jobId: jobId,
      _created: new Date().toISOString()
    };
    return {
      qitem: qitem,
      topic: topic
    };
  }
  var getRange = function getRange(params, sess) {
    // TODO: enqueue from query, indexes
    debug("getRange", util.inspect(params, {
      depth: 5
    }));
    var coll = getCollection(params.collection);
    // options
    var jobId = nanoid();
    var preResult = [];
    var filter = null;
    var useIndex = null;
    var startIndex = null;
    var endIndex = null;
    var limit = Number.MAX_SAFE_INTEGER;
    var offset = 0;
    var reverse = null;
    var hints = {};
    var onlyCount = false;
    var enqueue = null;
    var queueCount = 0;

    // Override params default
    if (params.optionsJSON.filter) {
      debug("Filter", params.optionsJSON.filter);
      filter = sift(params.optionsJSON.filter);
      debug("Filter", filter.toString());
    }
    if (params.optionsJSON.useIndex) {
      useIndex = params.optionsJSON.useIndex;
    }
    if (params.optionsJSON.startIndex) {
      startIndex = createIndexKey(params.optionsJSON.startIndex);
    }
    if (params.optionsJSON.endIndex) {
      endIndex = createIndexKey(params.optionsJSON.endIndex);
    }
    if (params.optionsJSON.limit) {
      limit = params.optionsJSON.limit;
    }
    if (params.optionsJSON.offset) {
      offset = params.optionsJSON.offset;
    }
    if (params.optionsJSON.reverse) {
      reverse = params.optionsJSON.reverse;
    }
    if (params.optionsJSON.hints && lodash.isObject(params.optionsJSON.hints)) {
      hints = params.optionsJSON.hints;
      onlyCount = hints.$onlycount || hints.$count || hints.count || false;
    }
    if (params.optionsJSON.enqueue && lodash.isString(params.optionsJSON.enqueue)) {
      enqueue = params.optionsJSON.enqueue;
    }
    var ee = new emitter();
    setTimeout(function () {
      if (useIndex) {
        var idxcoll = getCollection("_idx_".concat(params.collection, "_by_").concat(useIndex));
        var openStart = true;
        var openEnd = true;
        lodash.each(idxcoll, function (idxdata, key) {
          var data = coll[idxdata];
          if (!data) {
            // data is removed from main index, remove from this index
            delete idxcoll[key];
          }
          if (startIndex !== null) {
            if (key.startsWith(startIndex) === false) {
              //debug("Jump", key, startIndex)
              openStart = false; // skip
            } else {
              openStart = true;
            }
          }
          if (filter !== null) {
            if (filter(data)) {
              if (openStart && openEnd) {
                if (enqueue) {
                  var _createQueuePayload = createQueuePayload(key, enqueue, data, jobId, hints),
                    qitem = _createQueuePayload.qitem,
                    topic = _createQueuePayload.topic;
                  addToQueue(topic, qitem);
                  queueCount++;
                } else {
                  preResult.push(pickAndOmit(data, hints));
                }
              }
            }
          } else {
            if (openStart && openEnd) {
              if (enqueue) {
                var _createQueuePayload2 = createQueuePayload(key, enqueue, data, jobId, hints),
                  _qitem = _createQueuePayload2.qitem,
                  _topic = _createQueuePayload2.topic;
                addToQueue(_topic, _qitem);
                queueCount++;
              } else {
                preResult.push(pickAndOmit(data, hints));
              }
            }
          }
        });
      } else {
        lodash.each(coll, function (data, key) {
          if (filter !== null) {
            if (filter(data)) {
              if (enqueue) {
                var _createQueuePayload3 = createQueuePayload(key, enqueue, data, jobId, hints),
                  qitem = _createQueuePayload3.qitem,
                  topic = _createQueuePayload3.topic;
                addToQueue(topic, qitem);
                queueCount++;
              } else {
                preResult.push(pickAndOmit(data, hints));
              }
            }
          } else {
            if (enqueue) {
              var _createQueuePayload4 = createQueuePayload(key, enqueue, data, jobId, hints),
                _qitem2 = _createQueuePayload4.qitem,
                _topic2 = _createQueuePayload4.topic;
              addToQueue(_topic2, _qitem2);
              queueCount++;
            } else {
              preResult.push(pickAndOmit(data, hints));
            }
          }
        });
      }
      if (reverse) {
        preResult = lodash.reverse(preResult);
      }
      debug("Result length1", preResult.length);
      preResult = preResult.slice(offset, limit);
      debug("Result length", preResult.length);
      if (onlyCount) {
        ee.emit("data", {
          "data": {
            "count": preResult.length
          }
        });
      } else {
        if (enqueue) {
          ee.emit('data', {
            "data": {
              "count": queueCount,
              "jobId": jobId
            }
          });
        } else {
          lodash.each(preResult, function (dataItem) {
            ee.emit('data', {
              "data": dataItem
            });
          });
        }
      }
      ee.emit('end');
    }, 1);
    return ee;
  };
  var getOne = function getOne(params, sess, cb) {
    //getOne({ collection: coll, key: key, optionsJSON: j(opts) }, sess, (err, res) => 
    var coll = getCollection(params.collection);
    if (!coll[params.key]) {
      cb('Not found');
    } else {
      cb(null, {
        "data": coll[params.key]
      });
    }
  };
  var saveDocuments = function saveDocuments(params, sess, cb) {
    debug("saveDocuments", params);
    // client.save({ collection: coll, payloadJSON: j(doc) }, sess, (err, saveres) => {
    var coll = getCollection(params.collection);
    var data = params.payloadJSON;
    if (Array.isArray(data)) {
      lodash.each(data, function (item) {
        if (isObject(item) === false) throw new Error("Not an object: " + item);
        if (!item._id) {
          item._id = toHex(Date.now()) + '-' + nanoid();
        }
        if (item._ttl) {
          item._archiveAt = nowInMillis + parseInt(item._ttl);
        }
        coll[item._id] = item;
      });
    } else {
      if (isObject(data) === false) throw new Error("Not an object: " + _typeof(data));
      if (!data._id) {
        data._id = toHex(Date.now()) + '-' + nanoid();
      }
      if (data._ttl) {
        data._archiveAt = nowInMillis + parseInt(data._ttl);
      }
      coll[data._id] = data;
    }
    saveDB();
    updateIndexes(params.collection, ensureIndex);
    cb(null, {
      "data": data
    });
  };
  var updateOne = function updateOne(params, sess, cb) {
    var coll = getCollection(params.collection);
    var document = params.documentJSON;
    debug('updateOne', params);
    var specialOperator = document.$inc || document.$pullAll || document.$pull || document.$push || document.$addToSet || document.$set || document.$pop || document.$unset || document.$rename || document.$dropall;
    var data = coll[params.id];
    if (!data) {
      return cb("Not found");
    }
    if (specialOperator) {
      applyUpdateOperator(document, data, function (e, s) {
        if (e) debug(e, s);
      });
    } else {
      lodash.merge(data, document);
    }
    coll[params.id] = data;
    saveDB();
    updateIndexes(params.collection, ensureIndex);
    cb(null, {
      "data": data
    });
  };
  var replaceOne = function replaceOne(params, sess, cb) {
    var coll = getCollection(params.collection);
    var document = params.documentJSON;
    debug('replaceOne', params);
    var specialOperator = document.$inc || document.$pullAll || document.$pull || document.$push || document.$addToSet || document.$set || document.$pop || document.$unset || document.$rename || document.$dropall;
    var data = coll[params.id];
    if (!data) {
      return cb("Not found");
    }
    if (specialOperator) {
      applyUpdateOperator(document, data, function (e, s) {
        if (e) debug(e, s);
      });
    } else {
      coll[params.id] = document;
    }
    saveDB();
    updateIndexes(params.collection, ensureIndex);
    cb(null, {
      "data": coll[params.id]
    });
  };
  var updateMany = function updateMany(params, sess, cb) {
    //client.updateMany({ collection: coll, documentJSON: j(doc), optionsJSON: j(option) }, sess, (err, res) => {
    var coll = getCollection(params.collection);
    var document = params.documentJSON;
    var filter = null;
    if (params.optionsJSON.filter) {
      filter = sift(params.optionsJSON.filter);
    }
    debug('updateMany', params);
    var specialOperator = document.$inc || document.$pullAll || document.$pull || document.$push || document.$addToSet || document.$set || document.$pop || document.$unset || document.$rename || document.$dropall;
    var count = 0;
    lodash.each(coll, function (data, key) {
      if (filter) {
        if (filter(data)) {
          if (specialOperator) {
            applyUpdateOperator(document, data, function (e, s) {
              if (e) debug(e, s);
            });
            ++count;
          } else {
            coll[key] = lodash.merge(data, document);
            ++count;
          }
        }
      } else {
        if (specialOperator) {
          applyUpdateOperator(document, data, function (e, s) {
            if (e) debug(e, s);
          });
          ++count;
        } else {
          coll[key] = lodash.merge(data, document);
          ++count;
        }
      }
    });
    saveDB();
    updateIndexes(params.collection, ensureIndex);
    cb(null, {
      "data": {
        "count": count
      }
    });
  };
  var replaceMany = function replaceMany(params, sess, cb) {
    var coll = getCollection(params.collection);
    var document = params.documentJSON;
    var filter = null;
    if (params.optionsJSON.filter) {
      filter = sift(params.optionsJSON.filter);
    }
    debug('replaceMany', params);
    var count = 0;
    lodash.each(coll, function (data, key) {
      if (filter) {
        if (filter(data)) {
          coll[key] = lodash.cloneDeep(document);
          coll[key]._id = key;
          ++count;
        }
      } else {
        coll[key] = lodash.cloneDeep(document);
        coll[key]._id = key;
        ++count;
      }
      debug("data", data);
    });
    saveDB();
    updateIndexes(params.collection);
    cb(null, {
      "data": {
        "count": count
      }
    });
  };
  var removeOne = function removeOne(params, sess, cb) {
    //removeOne({ collection: coll, id: ID, optionsJSON: option }, sess, (err, res) => {
    var coll = getCollection(params.collection);
    if (!coll[params.id]) {
      return cb("Not found");
    }
    delete coll[params.id];
    saveDB();
    cb(null, {
      "data": params.id
    });
  };
  var removeMany = function removeMany(params, sess, cb) {
    var coll = getCollection(params.collection);
    var document = params.documentJSON;
    var filter = null;
    if (params.optionsJSON.filter) {
      filter = sift(params.optionsJSON.filter);
    }
    debug('removeMany', params);
    var count = 0;
    lodash.each(coll, function (data, key) {
      if (filter) {
        if (filter(data)) {
          delete coll[key];
          ++count;
        }
      } else {
        delete coll[key];
        ++count;
      }
    });
    saveDB();
    cb(null, {
      "data": {
        "count": count
      }
    });
  };
  var ensureIndex = function ensureIndex(params, sess, cb) {
    var ee = new emitter();
    var dataColl = getCollection(params.collection);
    var index = params.index;
    if (Array.isArray(index) === false) {
      // always arrays of indexes
      index = [index];
    }
    var metaColl = getCollection('_dbmetadata');
    if (!metaColl['indexes']) {
      lodash.set(metaColl, ['indexes', params.collection], {});
    }
    debug("Metacoll", metaColl);
    setTimeout(function () {
      lodash.each(index, function (aindex) {
        var unsorted = [];
        lodash.each(dataColl, function (dataItem, key) {
          debug("Idx key", key, dataItem[aindex]);
          var idxKey = createIndexKey(dataItem[aindex]);
          unsorted.push({
            key: "".concat(idxKey, "_").concat(key),
            val: key
          });
        });
        var sorted = sort(unsorted).asc('key');
        var coll = getCollection("_idx_".concat(params.collection, "_by_").concat(aindex));
        debug("start", sorted);
        // reset index
        lodash.each(coll, function (v, k) {
          delete coll[k];
        });
        lodash.each(sorted, function (v, k) {
          coll["".concat(v.key)] = v.val;
        });
        debug('end', coll);
        lodash.set(metaColl, ['indexes', params.collection, aindex], 'string');
        debug("Metacoll", metaColl, aindex);
      });
      saveDB();
      ee.emit('end', {
        "data": index
      });
    }, 100);
    return ee;
  };
  var kvset = function kvset(params, sess, cb) {
    // client.kvset({ key: key, value: value, optionsJSON: j(options || {}) }, sess, (err, setres) => {
    var keyspace = params.optionsJSON && params.optionsJSON.keyspace ? "_kvspace" + params.optionsJSON.keyspace : "_defaultkvspace";
    params.optionsJSON.keyspace = keyspace;
    //debug("KVSET", key, value, optionsJSON, keyspace);
    var coll = getCollection(keyspace);
    var jsonval = {
      "val": params.value
    };
    if (params.optionsJSON.ttl) {
      jsonval._archiveAt = Date.now() + parseInt(params.optionsJSON.ttl);
    }
    coll[params.key] = jsonval;
    cb(null, {
      data: jsonval.val
    });
    saveDB();
  };
  var kvget = function kvget(params, sess, cb) {
    var keyspace = params.optionsJSON && params.optionsJSON.keyspace ? "_kvspace" + params.optionsJSON.keyspace : "_defaultkvspace";
    params.optionsJSON.keyspace = keyspace;
    var coll = getCollection(keyspace);
    var jsonval = coll[params.key];
    if (!filterTTL(jsonval, Date.now(), keyspace, params.key)) {
      cb(null, {
        data: jsonval.val
      });
    } else {
      debug("KVGET data if filtered with TTL");
      cb({
        error: "no data for key"
      });
      saveDB();
    }
  };
  var kvgetAll = function kvgetAll(params, sess, cb) {
    var ee = new emitter();
    var keyspace = params.optionsJSON && params.optionsJSON.keyspace ? "_kvspace" + params.optionsJSON.keyspace : "_defaultkvspace";
    var onlykeys = params.optionsJSON.onlykeys && params.optionsJSON.onlykeys === true;
    setTimeout(function () {
      var coll = getCollection(keyspace);
      var result = [];
      lodash.each(coll, function (value, key) {
        if (!filterTTL(value, Date.now(), keyspace, key)) {
          if (key.startsWith(params.key)) {
            result.push({
              "key": key,
              "val": value.val
            });
          }
        }
      });
      saveDB();
      if (params.optionsJSON.reverse) {
        result = sort(result).desc('key');
      }
      lodash.each(result, function (item) {
        debug("Emit", item);
        ee.emit('data', item);
      });
      ee.emit('end');
    }, 1);
    return ee;
  };
  var kvdelete = function kvdelete(params, sess, cb) {
    debug(params);
    var keyspace = params.optionsJSON && params.optionsJSON.keyspace ? "_kvspace" + params.optionsJSON.keyspace : "_defaultkvspace";
    debug(keyspace);
    var coll = getCollection(keyspace);
    var jsonval = coll[params.key];
    cb(null, {
      data: jsonval
    });
    delete coll[params.key];
    saveDB();
  };
  var kvdeleteAll = function kvdeleteAll(params, sess, cb) {
    var keyspace = params.optionsJSON && params.optionsJSON.keyspace ? "_kvspace" + params.optionsJSON.keyspace : "_defaultkvspace";
    var onlykeys = params.optionsJSON.onlykeys && params.optionsJSON.onlykeys === true;
    var coll = getCollection(keyspace);
    var count = 0;
    lodash.each(coll, function (value, key) {
      if (key.startsWith(params.key)) {
        count++;
        debug("delAll one", key, value);
        delete coll[key];
      }
    });
    cb(null, {
      data: count
    });
    saveDB();
  };
  var kvincr = function kvincr(params, sess, cb) {
    var keyspace = params.optionsJSON && params.optionsJSON.keyspace ? "_kvspace" + params.optionsJSON.keyspace : "_defaultkvspace";
    var ttl = params.optionsJSON && params.optionsJSON.ttl ? params.optionsJSON.ttl : null;
    var coll = getCollection(keyspace);
    var data = coll[params.key];
    if (data === undefined) {
      data = {
        val: 0
      };
      debug("Inc init val", params.key, data);
    }
    if (filterTTL(data, Date.now(), keyspace, params.key)) {
      // removed old
      var jsonval = {
        "val": params.num
      };
      if (ttl) {
        jsonval._archiveAt = Date.now() + parseInt(ttl);
      }
      debug("KVINCR new", jsonval);
      coll[params.key] = jsonval;
    } else {
      // keep
      data.val = parseInt(data.val) || 0;
      data.val += params.num;
      // renew ttl if any
      if (ttl) {
        data._archiveAt = Date.now() + parseInt(ttl);
      }
      debug("Put incr", data);
      coll[params.key] = data;
    }
    cb(null, {
      "data": data
    });
    saveDB();
  };
  var kvdecr = function kvdecr(params, sess, cb) {
    params.num *= -1;
    kvincr(params, sess, cb);
  };
  var enqueue = function enqueue(topic, item) {
    return addToQueue(topic, item);
  };
  return {
    "getRange": getRange,
    "getOne": getOne,
    "save": saveDocuments,
    "getMany": getRange,
    "updateOne": updateOne,
    "replaceOne": replaceOne,
    "updateMany": updateMany,
    "replaceMany": replaceMany,
    "removeOne": removeOne,
    "removeMany": removeMany,
    "ensureIndex": ensureIndex,
    "kvset": kvset,
    "kvget": kvget,
    "kvgetall": kvgetAll,
    "kvdelete": kvdelete,
    "kvdeleteall": kvdeleteAll,
    "kvincr": kvincr,
    "kvdecr": kvdecr,
    "enqueue": enqueue
    /* 
    "getKeys": getKeys,
    "getValues": getValues,        
    "rebuildIndex": rebuildIndex,
    "removeIndex": removeIndex,
    "update": updateDocuments,
    "delete": deleteDocuments,*/
  };
};

exports.DB = _DB;