"use strict";

var lodash = require('lodash');
var debug = require('debug')('operators');
exports.pickAndOmit = function (obj, h) {
  var f = h['$fields'] || h['fields'];
  if (!f) {
    //debug("Missing hint")
    return obj;
  }
  obj = lodash.omitBy(obj, function (val, key) {
    if (f[key] === 0) {
      return true;
    }
  });
  var pick = lodash.pickBy(obj, function (val, key) {
    if (f[key] === 1) {
      return true;
    }
  });
  if (lodash.isEmpty(pick)) {
    return obj;
  } else {
    return pick;
  }
};
exports.myop = {};
exports.applySetOperator = function (document, vitem, cb) {
  //console.log("$SET OP")
  lodash.assign(vitem, document['$set']);
  cb(null, vitem);
};

/*
    $unset update operator
*/
exports.applyUnSetOperator = function (document, vitem, cb) {
  var unsets = lodash.isObject(document['$unset']) ? document['$unset'] : [document['$unset']];
  lodash.each(unsets, function (field, key) {
    debug("$UNSET item", field);
    delete vitem[field];
  });
  cb(null, vitem);
};

/*
    $rename update operator
*/
exports.applyRenameOperator = function (document, vitem, cb) {
  //debug("$RENAME OP before", document, vitem);
  if (lodash.isObject(document['$rename']) && !lodash.isArray(document['$rename'])) {
    lodash.each(document['$rename'], function (field, key) {
      var val = lodash.clone(vitem[key]);
      delete vitem[key];
      vitem[field] = val;
    });
  } else {
    return cb("Illegal $rename, not an Object", document['$rename']);
  }
  debug("$RENAME OP after", document, vitem);
  cb(null, vitem);
};

/*
    $pull update operator
*/
exports.applyPullOperator = function (document, vitem, cb) {
  lodash.each(lodash.toPairs(document['$pull']), function (pullOp) {
    debug("$pull OP", pullOp);
    if (vitem[pullOp[0]]) {
      //console.log("Pull",vitem[pullOp[0]])
      //console.log("Match", pullOp[1]);
      if (lodash.isObject(pullOp[1])) {
        //console.log("Complex pull",vitem[pullOp[0]], pullOp[1]);
        lodash.pullAllWith(vitem[pullOp[0]], [pullOp[1]], function (comp) {
          //console.log("comp", comp);
          return lodash.isEqual(comp, pullOp[1]);
        });
      } else {
        //console.log("Simple pull");
        lodash.pull(vitem[pullOp[0]], pullOp[1]);
      }
      //console.log("After pull", vitem);
    } else {
      console.error("Error", vitem, pullOp[0]);
      cb("No operand pull " + vitem[pullOp[0]], null);
    }
  });
  cb(null, vitem);
};

/*
    $pullAll update operator
*/
exports.applyPullAllOperator = function (document, vitem, cb) {
  // the same
  exports.applyPullOperator(document, vitem, cb);
};

/*
    $push update operator
*/
exports.applyPushOperator = function (document, vitem, cb) {
  lodash.each(lodash.toPairs(document['$push']), function (pushOp) {
    ////console.log("$push OP", pushOp);
    if (lodash.isObject(pushOp)) {
      if (vitem[pushOp[0]] && vitem[pushOp[0]].push) {
        vitem[pushOp[0]].push(pushOp[1]);
      } else {
        vitem[pushOp[0]] = [pushOp[1]];
      }
    } else {
      //
      cb("Cannot $push on a none Object", null);
      throw Error("Cannot $push with a string");
    }
  });
  cb(null, vitem);
};

/*
    $addToSet update operator.
    Push uniq objects to array.
*/
exports.applyAddToSetOperator = function (document, vitem, cb) {
  lodash.each(lodash.toPairs(document['$addToSet']), function (pushOp) {
    debug("$addToSet OP", vitem, pushOp);
    if (lodash.isObject(pushOp)) {
      var exists = lodash.find(vitem[pushOp[0]], function (o) {
        return lodash.isEqual(o, pushOp[1]);
      });
      if (!exists) {
        if (vitem[pushOp[0]] && vitem[pushOp[0]].push) {
          debug("1", pushOp[0]);
          vitem[pushOp[0]].push(pushOp[1]);
          // item exists
        } else {
          vitem[pushOp[0]] = [pushOp[1]];
        }
      }
    } else {
      //
      cb("Cannot $push on a none Object", null);
      throw Error("Cannot $push with a string");
    }
  });
  cb(null, vitem);
};

/*
    $inc update operator
*/
exports.applyIncOperator = function (document, vitem, cb) {
  lodash.each(lodash.toPairs(document['$inc']), function (incOp) {
    //console.log("$inc OP", incOp);
    if (lodash.isNumber(incOp[1])) {
      if (vitem[incOp[0]]) {
        vitem[incOp[0]] += lodash.parseInt(incOp[1]);
      } else {
        vitem[incOp[0]] = lodash.parseInt(incOp[1]);
      }
    } else {
      //
      cb("Cannot $inc with a string", null);
      throw Error("Cannot $inc with a string");
    }
  });
  cb(null, vitem);
};

/*
    $pop update operator
*/
exports.applyPopOperator = function (document, vitem, cb) {
  var pops = lodash.isObject(document['$pop']) ? document['$pop'] : [document['$pop']];
  lodash.each(pops, function (popOp) {
    debug("$pop OP", popOp);
    if (vitem[popOp] && vitem[popOp].pop) {
      vitem[popOp].pop();
    } else {
      cb("No pop on", vitem);
    }
  });
  cb(null, vitem);
};